/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "siworin-config.h"
#include "siworin-match.h"
#include "siworin-displ.h"

/* See also rudhar.com/sfreview/siworin/siworin19.htm */

/* 5 and 6 November 2022:
   Fixed a bug. For example when three words matched, adding
   an extra search word, occurring in the same file but too
   far away, made the whole match disappear because of the
   too great distance. But of course even though the 4-word
   match does not hold, the 3-word match still does!

   5 September 2024:
   Measures for special case when two or more of the search
   words are indentical. Dutch example that actually occurs
   on my website rudhar.com: dat dat dat.

   23 September 2024:
   Added a configurable limit to the number of results to
   display.
 */

typedef double rank_t;
static rank_t calc_ranking (
   struct hfile_wordvector_s *matches, int numelts_hfile_wordvector);
static int comparrank (const void *a, const void *b);

static unsigned char *hitstab;
static int numhits;
static size_t binrecsize;
static rank_t ranking_numwords;
static rank_t ranking_wordfreq;
static rank_t ranking_worddist;

void siworin_ranking_init (int numelts_hfile_wordvector)
{
   hitstab = NULL;
   numhits = 0;
   binrecsize = sizeof (rank_t) +
      sizeof (struct hfile_wordvector_s) * numelts_hfile_wordvector;

   ranking_numwords = atof(siworin_config_value(SIWORIN_RANKING_NUMWORDS));
   ranking_wordfreq = atof(siworin_config_value(SIWORIN_RANKING_WORDFREQ));
   ranking_worddist = atof(siworin_config_value(SIWORIN_RANKING_WORDDIST));
}

int siworin_ranking_add1hit (
   struct hfile_wordvector_s *matches, int numelts_hfile_wordvector)
{
   rank_t rank = calc_ranking(matches, numelts_hfile_wordvector);

   hitstab = realloc(hitstab, ++numhits * binrecsize);
   if (!hitstab)
      return 1;

   /* Copy the data to a sortable binary table in core */
   memcpy(hitstab + (numhits-1) * binrecsize, (unsigned char *)&rank,
          sizeof (rank_t));
   for (int i = 0; i < numelts_hfile_wordvector; i++)
   {
      memcpy(hitstab + (numhits-1) * binrecsize +
         sizeof (rank_t) + i * sizeof matches[0],
         (unsigned char *)(matches + i),
         sizeof matches[0]);
   }

   return 0;
}

void siworin_ranking_finalise (int numelts_hfile_wordvector)
{
   qsort(hitstab, numhits, binrecsize, comparrank);

   /* Read the binary table and display search results. */
   int to_display = atoi(siworin_config_value(SIWORIN_MAX_RESULTS_TO_DISPLAY));
   if (to_display > 0 && to_display < numhits)
      numhits = to_display;

   for (int i = 0; i < numhits; i++)
   {
      struct hfile_wordvector_s *thisone;

      /* Use the data from the binary table, that was built up in
         function siworin_ranking_add1hit */
      thisone = (struct hfile_wordvector_s *)
         (hitstab + i * binrecsize + sizeof (rank_t));

      #ifdef DEBUG
      fprintf(stdout, "<p>%s line %d: Ranking is %15.12f</p>\n",
         __FILE__, __LINE__,
         (double)*(rank_t *)(hitstab + i * binrecsize));
      #endif
      siworin_displaymatches(thisone, numelts_hfile_wordvector);
   }

   free(hitstab);
   hitstab = NULL;
   numhits = 0;
}

static rank_t calc_ranking (
   struct hfile_wordvector_s *matches, int numelts_hfile_wordvector)
{
   rank_t rank = 0;
   rank_t toadd;
   int i;
   unsigned long prev_pathline = matches[0].vec[matches[0].currind].pathline;
   long prev_offseth  = matches[0].vec[matches[0].currind].offseth;
   int maxim_dista = atoi(siworin_config_value(SIWORIN_MAXIM_DISTA));

   for (i = 0;
        i < numelts_hfile_wordvector &&
        matches[i].currind < matches[i].numelts_vec &&
        matches[i].vec[matches[i].currind].pathline == prev_pathline &&
        (long)
        matches[i].vec[matches[i].currind].offseth - prev_offseth <= maxim_dista;
        i++)
   {
      /* A hit on a rarer word is worth more than a hit
         on a common, often occurring word. */
      toadd = ranking_wordfreq / matches[i].numelts_vec;
      rank += toadd;
      #ifdef DEBUG
      fprintf(stdout,
         "<p>%s line %d:<br>Ranking is %15.12f, just added %15.12f</p>\n",
         __FILE__, __LINE__, rank, toadd);
      #endif

      /* Matching words that are close to each other rank
         higher than distant words. */
      if (i > 0)
      {
         int dist =
            matches[i  ].vec[matches[i  ].currind].offseth -
            matches[i-1].vec[matches[i-1].currind].offseth;

         /* The distance could only be zero in case one or
            more search words are identical. But function
            SpecialCaseEqualSrchwrds in module siworin-match.c
            arranges that each of the search words then matches
            a different word in the file under consideration.

            However, safety first, never divide by zero,
            whatever happens. */
         if (dist)
         {
            toadd = ranking_worddist / dist;
            rank += toadd;
            #ifdef DEBUG
            fprintf(stdout,
               "<p>%s line %d:<br>Ranking is %15.12f, just added %15.12f</p>\n",
               __FILE__, __LINE__, rank, toadd);
            #endif
         }
      }
   }
   /* More matching words means there is a better match. */
   toadd = ranking_numwords * i;
   rank += toadd;
   #ifdef DEBUG
   fprintf(stdout,
      "<p>%s line %d:<br>Ranking is %15.12f, just added %15.12f</p>\n",
      __FILE__, __LINE__, rank, toadd);
   #endif

   /* To conclude, add a fraction of the line number of the
      file path, and of the last offset, so otherwise identical
      rankings are made different. This makes sorting results
      deterministic, that might otherwise depend on implementation
      details of the sorting algorithm.

      This makes life easier when doing comparison tests with
      different compilers, libraries and Unix platforms. Any
      difference is now suspect. Results should be identical
      across all. */
   toadd =
      matches[0].vec[matches[0].currind].pathline / 1000000.0 +
      matches[0].vec[matches[0].currind].offseth  / 1000000.0 / 1000.0;
   rank += toadd;
   #ifdef DEBUG
   fprintf(stdout,
      "<p>%s line %d:<br>Ranking is %15.12f, just added %15.12f</p>\n",
      __FILE__, __LINE__, rank, toadd);
   #endif

   return rank;
}

static int comparrank (const void *a, const void *b)
{
   /* The rank is at the start of the binary record, so
      casting the pointer to the record to type rank_t
      gives a pointer to the rank. Very simple. */
   rank_t dif = *(rank_t *)b - *(rank_t *)a;
   rank_t zer = 0.0;

   return dif > zer ? 1 : -1;
}
